from typing import Dict, Any
import logging
from pyspark.sql.functions import col, when, lit, to_date, date_format, date_add
from prefect import task

from Enum.DataTypeEnum import DataTypeEnum
from Enum.DatabaseTypeEnum import DatabaseTypeEnum
from Enum.InputTypeEnum import InputTypeEnum
from Utils.SparkUtils import createSession
from Input.Source import Input

logger = logging.getLogger()


class ETLProcess:

    def __init__(self, config: Dict[str, Any]) -> None:
        self.conf = config
        self.identifier = self.conf["identifier"]
        self.session = None

        self.inputs = {}

    def init(self, spark_jars: Dict[str, str], source_type: InputTypeEnum = InputTypeEnum.BUCKET) -> None:
        self.session = createSession(self.identifier, spark_jars, source_type)

    @task
    def reader(self) -> None:
        try:
            inputs = self.conf["inputs"]
            input_type = inputs["type"]
            provider = inputs["params"]["provider"] if "provider" in inputs["params"].keys() else None
            for input_obj in inputs["data"]:
                identifier = input_obj["identifier"]
                params = {"identifier": identifier, "path": input_obj["path"], "type": input_obj["input_type"],
                          "separator": input_obj["separator"], "schema": input_obj["schema"]}
                current_input = Input(input_type, self.session, params, provider)
                # Caso especial de reemplazar "\t" con " "
                if identifier == "FACTURACION":
                    current_input.get_data(True)
                else:
                    current_input.get_data()
                self.inputs.update({identifier: current_input.data})
        except Exception as e:
            raise AssertionError(f"Error in function extrayendo data. Reader. {e}")

    @task
    def set_schema(self) -> None:
        try:
            inputs = self.conf["inputs"]
            for input_obj in inputs["data"]:
                identifier = input_obj["identifier"]
                schema = input_obj["schema"]
                input_schema = self.create_schema(schema)
                columns, schema = input_schema["columns"], input_schema["schema"]
                self.inputs[identifier] = self.inputs[identifier].select(*columns)
                for column, datatype in schema:
                    self.inputs[identifier] = self.inputs[identifier].withColumn(column, col(column).cast(datatype))
        except Exception as e:
            raise AssertionError(f"Error procesando información. Process. {e}")

    def create_schema(self, schema: Dict[str, str]) -> Dict[str, Any]:
        response = {}
        try:
            columns = list(schema.keys())
            structure = []
            for column in columns:
                field = (column, DataTypeEnum[schema[column]].value())
                structure.append(field)
            response.update({"columns": columns, "schema": structure})
        except Exception as e:
            logger.error(f"Error leyendo esquema para el insumo {self.identifier}. {e}")
        finally:
            return response

    @task
    def process_gross(self, identifier: str) -> bool:
        success = False
        try:
            self.inputs[identifier] = self.inputs[identifier].withColumn("AGENTE_COMISIONA", col("CONSULTOR_NK"))
            self.inputs[identifier] = self.inputs[identifier].withColumn('SEGMENTO',
                when((col('PLAN_NOMBRE').contains('Inter')) & (col('CLIENTE_NATURALEZA') == 'Persona Juridica'), 'B2B')
                .when(col('PLAN_NOMBRE').contains('Neg'), 'B2B')
                .when(col('SERVICIO').contains('Prepaid'), 'PREP').otherwise('B2C'))
            self.inputs[identifier] = self.inputs[identifier].withColumn("TIPO_CANAL", lit("DIRECT"))
            success = True
        except Exception as e:
            logger.error(f"Error transformando archivo gross. {e}")
        finally:
            return success

    @task
    def process_teams(self, identifier: str) -> bool:
        success = False
        try:
            self.inputs[identifier] = self.inputs[identifier].withColumn('SEGMENTO',
                 when(col('CLIENTE_NATURALEZA') == 'Persona Juridica', 'B2B')
                .otherwise('B2C'))
            self.inputs[identifier] = self.inputs[identifier].withColumn("TIPO_CANAL", lit("DIRECT"))
            success = True
        except Exception as e:
            raise AssertionError(f"Error transformando archivo equipo. {e}")
        finally:
            return success

    @task
    def process_facturacion(self, identifier: str) -> bool:
        success = False
        try:
            df = self.inputs[identifier]
            df = df.withColumn("fecha_vencimiento_fact", to_date(df["FECHA_VENCIMIENTO"], "dd/MM/yy"))

            df = df.withColumn("fecha_periodo_fact",
                               to_date(date_format(col("PERIODO_PROCESO_CODIGO"), "yyyyMM") + "01", "yyyyMMdd"))

            df = df.withColumn("FACTURA_VENCIDA",
                               when(date_add(col("fecha_periodo_fact"), 5) < col("fecha_vencimiento_fact"), lit("SI"))
                               .otherwise(lit("NO")))

            self.inputs[identifier] = df
            success = True
        except Exception as e:
            logger.error(f"Error transformando archivo de facturacion. {e}")
        finally:
            return success

    @task
    def write(self, identifier: str, starroks_jdbc: str, starroks_fe: str, prev_status: bool = True,
              db_type: DatabaseTypeEnum = DatabaseTypeEnum.REDSHIFT, redshift_url: str = "", mysql_url: str = "") -> None:
        try:
            if db_type == DatabaseTypeEnum.REDSHIFT:
                self.inputs[identifier].coalesce(45).write \
                    .format("jdbc") \
                    .option("driver", "com.amazon.redshift.jdbc42.Driver") \
                    .option("url", redshift_url) \
                    .option("dbtable", identifier) \
                    .option("user", "awsuser") \
                    .option("password", "Awsuser123") \
                    .mode("append") \
                    .save()
            elif db_type == DatabaseTypeEnum.MYSQL:
                mysql_user = self.conf["mysql"]["user"]
                mysql_pass = self.conf["mysql"]["password"]
                self.inputs[identifier].write \
                  .format("jdbc") \
                  .option("driver", "com.mysql.cj.jdbc.Driver") \
                  .option("url", mysql_url) \
                  .option("dbtable", identifier) \
                  .option("user", mysql_user) \
                  .option("password", mysql_pass) \
                  .mode("append") \
                  .save()
            else:
                database = starroks_jdbc[starroks_jdbc.rfind("/")+1:]
                starroks_user = self.conf["starroks"]["user"]
                starroks_pass = self.conf["starroks"]["password"]

                self.inputs[identifier].write.format("starrocks") \
                    .option("starrocks.fe.http.url", starroks_fe) \
                    .option("starrocks.fe.jdbc.url", starroks_jdbc) \
                    .option("starrocks.table.identifier", database+"."+identifier) \
                    .option("starrocks.user", starroks_user) \
                    .option("starrocks.password", starroks_pass) \
                    .mode("append") \
                    .save()
        except Exception as e:
            logger.error(f"Error guardando resultados. {e}")
